package ee.temporalcorrelation;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

import ee.config.Config;
import ee.decomposition.Decompose;
import ee.input.file.FileFinder;
import ee.input.file.ReadFile;
import ee.session.FilteredSessionTotalFrequency;
import ee.session.SessionExtraction;
import ee.session.SessionFiltering;
import ee.termgroups.SimpleTermGroups;

public class TemporalCorrelationPreprocessing {
	private HashMap correlationMap;
	private int[] maxQueryTimeUnit;
	private int totalNumberOfQueries;
	private ArrayList correlationResultArray;
	private HashMap sessionQueryFrequencyMap;
	
	public class TimeUnit{
		String query;
		int frequency[];
		int totalFrequency;
		float averageFrequency;
		float standardDeviation;
		float preCalcCorrelationValues[];
		int numberOfZeroFrequencyies = 0;
		
		public TimeUnit(String query, int numberOfTimeUnits){
			this.query = query;
			this.frequency = new int[numberOfTimeUnits];
			preCalcCorrelationValues = new float[numberOfTimeUnits];
		}
	}
	
	public static class CorrelationTotalFrequencyComperator implements Comparator {
		public int compare(Object sf1, Object sf2) {
			TimeUnit sortInfo1 = (TimeUnit)sf1;
			TimeUnit sortInfo2 = (TimeUnit)sf2;
			
			if (sortInfo1.totalFrequency == sortInfo2.totalFrequency) {
				return 0;
			} else if (sortInfo1.totalFrequency > sortInfo2.totalFrequency) {
				return -1;
			} else {
				return 1;
			}
		}
	}
	
	public static class CorrelationValueComperator implements Comparator {
		public int compare(Object sf1, Object sf2) {
			CorrelationFields sortInfo1 = (CorrelationFields)sf1;
			CorrelationFields sortInfo2 = (CorrelationFields)sf2;
			
			if (sortInfo1.correlationValue == sortInfo2.correlationValue) {
				return 0;
			} else if (sortInfo1.correlationValue > sortInfo2.correlationValue) {
				return -1;
			} else {
				return 1;
			}
		}
	}
	
	public static class CorrelationInfoComperator implements Comparator {
		public int compare(Object sf1, Object sf2) {
			CorrelationInfo sortInfo1 = (CorrelationInfo)sf1;
			CorrelationInfo sortInfo2 = (CorrelationInfo)sf2;
			
			if (sortInfo1.totalQueryFrequency == sortInfo2.totalQueryFrequency) {
				return 0;
			} else if (sortInfo1.totalQueryFrequency > sortInfo2.totalQueryFrequency) {
				return -1;
			} else {
				return 1;
			}
		}
	}
	
	public static class CorrelationInfo{
		public String query;
		public String mergeSegment;
		public ArrayList correlationList;
		public int totalQueryFrequency;
		public CorrelationInfo(String query){
			this.query = query;
			correlationList = new ArrayList();
		}
	}
	
	public static class CorrelationFields{
		public String query;
		public float correlationValue;
		public int frequency;
		public CorrelationFields(String query, float correlationValue, int frequency){
			this.query = query;
			this.correlationValue = correlationValue;
			this.frequency = frequency;
		}
		
		public CorrelationFields(String query, float correlationValue){
			this.query = query;
			this.correlationValue = correlationValue;
		}
	}

	public void preprocess() throws IOException{
		input();
		precalculate();
		WriteCorrelationData.writeFrequencyStatistics(correlationMap);
		calculateCorrelations();
		WriteCorrelationData.writeCorrelations(correlationResultArray, "correlation_values_test");
	}
	
	public void processCorrelatinsWithFiltering() throws IOException{
		SessionExtraction.findSessions();
		Decompose decompose = new Decompose();
		SessionFiltering sf = new SessionFiltering(decompose);
		sessionQueryFrequencyMap = FilteredSessionTotalFrequency.createTotalFrequencySessionMap();
		
		input();
		precalculate();
		calculateCorrelations();
		System.out.println("Calculated correlations");
		
		FilterCorrelations fc = new FilterCorrelations(correlationResultArray, decompose, sf);
		
		ArrayList filteredCorrelations = fc.filterCorrelationsWithSessionQueriesAndSegmenation();
		ArrayList mergedFilteredCorrelations = FilterCorrelations.mergeCorrelationGroups(filteredCorrelations);
		
		//WriteCorrelationData.writeCorrelations(correlationResultArray, "correlation_values_" + Config.createFileInfoString());
		//WriteCorrelationData.writeCorrelations(filteredCorrelations, "query_filtered_correlations" + Config.createFileInfoString());
		//WriteCorrelationData.writeCorrelations(mergedFilteredCorrelations, "segment_merged_query_filtered_correlations" + Config.createFileInfoString());

		WriteCorrelationData.writeCorrelations(fc.trimCorrelationGroups(mergedFilteredCorrelations), "trimmed_segment_merged_query_filtered_correlations" + Config.createFileInfoString());
		
		SimpleTermGroups.createGroupsWithSegments(mergedFilteredCorrelations, decompose);
		SimpleTermGroups.createGroupsByExtraction(mergedFilteredCorrelations, decompose);
		
		//WriteCorrelationData.writeCorrelations(TermGroups.createGroups(mergedFilteredCorrelations, decompose), "term_groups" + createFileInfoString());
		//writeCorrelations(filterCorrelationsWithSessionTerms(sf), "term_filtered_correlations" + createFileInfoString());
	}
	
	private void input() throws IOException{
		ArrayList fileList = FileFinder.getAllFiles(Config.PATHS_PREPROCESSES_RAW_LOG_PATH, "frequency_access");
		
		correlationMap = new HashMap();
		maxQueryTimeUnit = new int[Config.CORRELATION_TIME_DURATION/Config.CORRELATION_TIME_UNIT];
		int counter = 0;
		
		for(int i = Config.CORRELATION_SKIP_FROM_START; i < Config.CORRELATION_SKIP_FROM_START + Config.CORRELATION_TIME_DURATION;){
			for(int j = 0; j < Config.CORRELATION_TIME_UNIT; j++){
				System.out.println(i);
				File f = (File)fileList.get(i);
				ReadFile rf = new ReadFile(f);
				String tempString = "";
				while((tempString = rf.getNextLine()) != null){					
					tempString = tempString.replaceAll("\"", "").trim();
					
					StringTokenizer st = new StringTokenizer(tempString, "\t");
					if(st.countTokens() == 2){
						int frequency = Integer.parseInt(st.nextToken().trim());
						String query = st.nextToken().trim();
						
						if(correlationMap.containsKey(query)){
							TimeUnit tu = (TimeUnit)correlationMap.get(query);
							tu.frequency[counter] += frequency;
							tu.totalFrequency += frequency;
						} else {
							TimeUnit tu = new TimeUnit(query, Config.CORRELATION_TIME_DURATION/Config.CORRELATION_TIME_UNIT);
							tu.frequency[counter] += frequency;
							tu.totalFrequency += frequency;
							correlationMap.put(query, tu);
						}
					} else {
						StringTokenizer st2 = new StringTokenizer(tempString, " ");
						if(st2.countTokens() ==2){
							int value = Integer.parseInt(st2.nextToken());
							maxQueryTimeUnit[counter] += value;
							totalNumberOfQueries += value;
							System.out.println(maxQueryTimeUnit[counter]);
						}
					}
				}
				i++;
			}
			counter++;
		}	
	}
	
	private void precalculate(){
		
		Iterator it = correlationMap.entrySet().iterator();
		while(it.hasNext()){
			Map.Entry me = (Map.Entry)it.next();
			TimeUnit tu = (TimeUnit)me.getValue();
			
			tu.averageFrequency = (float)tu.totalFrequency / (float)maxQueryTimeUnit.length;
			
			float difference = 0;
			for(int j = 0; j < maxQueryTimeUnit.length; j++){
				difference += (tu.frequency[j]-tu.averageFrequency) * (tu.frequency[j]-tu.averageFrequency);
				if(tu.frequency[j] == 0){
					tu.numberOfZeroFrequencyies++;
				}
			}
			tu.standardDeviation = (float)Math.sqrt((double)difference / (double)maxQueryTimeUnit.length);
			
			for(int i = 0; i < maxQueryTimeUnit.length; i++){
				tu.preCalcCorrelationValues[i] = ((float)tu.frequency[i] - tu.averageFrequency) / tu.standardDeviation;	
			}
		}
	}
	
	private void calculateCorrelations(){
		
		correlationResultArray = new ArrayList();
		
		for(Iterator it1 = correlationMap.entrySet().iterator(); it1.hasNext(); ){
			Map.Entry me1 = (Map.Entry) it1.next();
			TimeUnit tu1 = (TimeUnit)me1.getValue();
			
			if(!sessionQueryFrequencyMap.containsKey(tu1.query) || ((Integer)sessionQueryFrequencyMap.get(tu1.query)).intValue() < Config.CORRELATION_MINIMUM_SESSION_QUERY_FREQUENCIES || tu1.numberOfZeroFrequencyies > Config.CORRELATION_NUMBER_OF_MAX_ZERO_FREQUENCIES){
				continue;
			}
			
			if(tu1.totalFrequency < Config.CORRELATION_MINIMUM_TOTAL_FREQUENCY || tu1.numberOfZeroFrequencyies > Config.CORRELATION_NUMBER_OF_MAX_ZERO_FREQUENCIES){
				continue;
			}
			
			CorrelationInfo ci = new CorrelationInfo(tu1.query);
			
			for(Iterator it2 = correlationMap.entrySet().iterator(); it2.hasNext(); ){
				Map.Entry me2 = (Map.Entry) it2.next();
				TimeUnit tu2 = (TimeUnit)me2.getValue();
				
				if(!sessionQueryFrequencyMap.containsKey(tu2.query) || ((Integer)sessionQueryFrequencyMap.get(tu2.query)).intValue() < Config.CORRELATION_MINIMUM_SESSION_QUERY_FREQUENCIES || tu1.numberOfZeroFrequencyies > Config.CORRELATION_NUMBER_OF_MAX_ZERO_FREQUENCIES){
					continue;
				}
				
				if(tu2.totalFrequency < Config.CORRELATION_MINIMUM_TOTAL_FREQUENCY || tu2.numberOfZeroFrequencyies > Config.CORRELATION_NUMBER_OF_MAX_ZERO_FREQUENCIES){
					continue;
				}

				float correlation = 0;
				for(int i = 0; i < tu1.frequency.length; i++){
					correlation += tu1.preCalcCorrelationValues[i] * tu2.preCalcCorrelationValues[i];
				}
				correlation = correlation / (float)tu1.frequency.length;
				
				if(correlation > Config.CORRELATION_COEFFICIENT){
					ci.correlationList.add(new CorrelationFields(tu2.query, correlation));
				}
			}
			correlationResultArray.add(ci);
		}
	}
	

	
	
	public static void main(String args[]) throws IOException{
		TemporalCorrelationPreprocessing tcp = new TemporalCorrelationPreprocessing();
//		Config.setCorrelationParameters(23, 1, 0, 10, 10);
//		tcp.processCorrelatinsWithFiltering();
//		Config.setCorrelationParameters(24, 1, 23, 10, 10);
//		tcp.processCorrelatinsWithFiltering();	
//		Config.setCorrelationParameters(24, 1, 47, 10, 10);
//		tcp.processCorrelatinsWithFiltering();
//		Config.setCorrelationParameters(12, 1, 11, 5, 10);
//		tcp.processCorrelatinsWithFiltering();
//		Config.setCorrelationParameters(48, 2, 0, 10, 25);
//		tcp.processCorrelatinsWithFiltering();
//		Config.setCorrelationParameters(69, 3, 0, 12, 40);
//		tcp.processCorrelatinsWithFiltering();
		Config.setCorrelationParameters(70, 5, 0, 6, 25, 3);
		tcp.processCorrelatinsWithFiltering();
	}
}
